# Copyright 2017 Department of Electrical and Computer Engineering, Cornell University. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# This is a package for locally private data transmission. 

import random
from pure_ldp.frequency_oracles.hadamard_response.internal.functions import *

#the Hadamard randamized responce when \epsilon < 1
class Hadamard_Rand_high_priv:
    def __init__(self, absz, pri_para, encode_acc = 0): # absz: alphabet size, pri_para: privacy parameter
        #set encode_acc = 1 to enable fast encoding by intializing hadamard matrix
        self.insz = absz #input alphabet size k
        self.outsz = int(math.pow(2,math.ceil(math.log(absz+1,2)))) #output alphabet size: smallest exponent of 2 which is bigger than k
        self.outbit = int(math.ceil(math.log(absz+1,2))) #output bit length
        self.exp = math.exp(pri_para) #privacy parameter
        self.pri_para = 1/(1+math.exp(pri_para)) #flipping probability to maintain local privacy
        #self.permute, self.reverse = Random_Permutation(absz) #Initialize random permutation
        
        self.ifencode_acc = encode_acc #whether to accelerate encoding process
        if encode_acc == 1:
            self.H = Hadarmard_init(self.outsz) # initialize Hadarmard matrix
            
    def parity_check(self,x,y): #check if the hadamard entry is one (return 0 for if the entry is 1)
        z = x&y #bit and
        z_bit = bin(z)[2:].zfill(self.outbit)
        check = 0
        for i in range(0,self.outbit): #check = \sum_i (x_i y_i) （mod 2）
            check = check^int(z_bit[i]) 
        return check
                                  
    def encode_symbol(self,in_symbol):  # encode a single symbol into a privatized version
        bitin = bin(int(in_symbol)+1)[2:].zfill(self.outbit) #map the input symbol x to x+1 since we are not using the first column of the matrix
        out1 = random.randint(0,math.pow(2,self.outbit)-1) #get a random number in the output alphabet as one potential output
        bitout1 = bin(out1)[2:].zfill(self.outbit)
        for i in range(0,self.outbit): #flip the bit of out1 corresponding to the left most bit in (in_symbol+1) which is one to get the other potential output
            if int(bitin[i]) == 1:
                out2 = out1 ^ (pow(2,self.outbit - i -1))
                break   
        #bitout2 = bin(out2)[2:].zfill(self.outbit)
        
        if self.ifencode_acc == 1:
            check = 1 - self.H[int(in_symbol)+1][out1]
        else:
            check = 0
            for i in range(0,self.outbit): # check if the Hadamard entry at position (in_symbol+1, out1) is one or not
                check = check^(int(bitout1[i])&int(bitin[i]))

        ra = random.random()
        if check == 0: # if it is one output out1 with prob (1-pri_para)
            if ra > self.pri_para:
                return out1
            else:
                return out2
        else: # else output out2 with prob (1-pri_para)
            if ra > self.pri_para:
                return out2
            else:
                return out1       
     
    def encode_string(self,in_list):  # encode string into a privatized string
        out_list = [self.encode_symbol(x) for x in in_list] # encode each symbol in the string
        return out_list
    
    def decode_string(self, out_list, iffast = 1, normalization = 0): # get the privatized string and learn the original distribution
        #normalization options: 0: clip and normalize(default)
        #                       1: simplex projection
        #                       else: no nomalization
        
        #iffast: 0 use fast hadamard transform time O(n  + k\log k)
        #        1 no fast hadamard tansform  time O(n + k^2)
        
        l = len(out_list) 
        count,edges = np.histogram(out_list,range(self.outsz+1))
        dist = count/float(l)
        
        if iffast == 1: #if we use fast hadamard transform
            dist_mid = FWHT_A(self.outsz, dist) #do fast hadamard transform to the frequency vector
            dist_S = (dist_mid[1:self.insz+1] + 1)/float(2) #get the frequencies of C_i
        else: #if we don't use fast hadamard transform
            num = [0]*self.insz
            for x in range(0,self.outsz):
            #print x
                for i in range(1, self.insz+1): #if output x is in C_i(locations in row i which is 1), add it to the count of C_i
                    if self.parity_check(i,x) == 0:
                        num[i-1] = num[i-1] + count[x]
            dist_S = np.array([float(x)/float(l) for x in num]) #get the frequencies of C_i
            
        dist = (2*dist_S*(1+self.exp)-(1+self.exp))/float(self.exp-1) #calculate the corresponding estimate for p_i
        if normalization == 0: 
            dist = probability_normalize(dist) #clip and normalize
        if normalization == 1:
            dist = project_probability_simplex(dist) #simplex projection
        
        return dist


#The Hadamard randomized response for all regimes (Modified Version)

class Hadamard_Rand_general:
    def __init__(self, absz, pri_para, encode_acc = 0): # absz: alphabet size, pri_para: privacy parameter
        self.insz = absz #input alphabet size k
        #self.outsz = int(math.pow(2,math.ceil(math.log(absz+1,2)))) #output alphabet size: smallest exponent of 2 which is bigger than k
        #self.outbit = int(math.ceil(math.log(absz+1,2))) #output bit length
        self.pri_para = 1/(1+math.exp(pri_para)) #flipping probability to maintain local privacy
        self.exp = math.exp(pri_para) #privacy parameter
        #self.initbit = int(math.floor(math.log(self.exp,2))) # number of bits indicating the block number 
        self.initbit = int(math.floor(math.log(min(2*absz,self.exp),2))) # number of bits indicating the block number
        self.part = int(math.pow(2,self.initbit)) #total number of blocks B
        self.tailbit = int(math.ceil(math.log(float(self.insz)/float(self.part)+1,2))) #number of bits indicating the location within a block
        self.partsz = int(math.pow(2,self.tailbit)) # size of each block b
        self.num_one = int(self.partsz/float(2))
        self.outbit = self.tailbit + self.initbit #total number of output bits
        self.partnum = int(math.ceil(float(self.insz)/float(self.partsz - 1)))
        self.outsz = int(self.partsz*self.partnum) # output alphabet size K
        self.permute, self.reverse = Random_Permutation(absz) #Initialize random permutation
        
        self.ifencode_acc = encode_acc #whether to accelerate encoding process
        if encode_acc == 1:
            self.H = Hadarmard_init(self.partsz) # initialize Hadarmard matrix
        
    def entry_check(self,x,y): #check if the reduced hadamard entry is one (return 0 for 1)
        x_bit = bin(x)[2:].zfill(self.outbit)
        y_bit = bin(y)[2:].zfill(self.outbit)
        for i in range(0,self.initbit): # check if they are in the same block, if not, return -1
            if x_bit[i] != y_bit[i]:
                return True
        check = False
        for i in range(self.initbit, self.outbit): #check whether the entry is one within the block
            check = check^(int(x_bit[i]) & int(y_bit[i]))
        return check
                                  
            
    def encode_symbol_rough(self,in_symbol):  # encode a single symbol into a privatized version 
        # we use coupling argument to do this
        part_index = int(in_symbol)//(self.partsz-1)
        part_pos = int(in_symbol)%(self.partsz-1)+1
        in_column = (part_index << self.tailbit) + part_pos #map the input x to the xth column with weight b/2
        #in_column = part_index * self.partsz + part_pos
        out1 = np.random.randint(0,self.outsz) #get a random number out1 in the output alphabet as a potential output
        ra = random.random()
        if ra < (2*self.part)/(2*self.part-1+self.exp): #with certain prob, output the same symbol as from uniform distribution
            return out1
        else:
            out_pos = out1 & (self.partsz - 1)
            #out_pos = out1 % self.partsz
            out1 =  out_pos + (part_index << self.tailbit) # map out1 to the same block as in_column while maintain the location within the block
            #out1 = out_pos + part_index*self.partsz
            if self.ifencode_acc == 1:
                check = self.H[part_pos][out_pos]
            else:
                check = 1 - self.entry_check(in_column,out1)

            if check == 0: #if location (in_column, out1) is one, output out1
                return out1
            else: #else flip bit at the left most location where bit representation of in_column is one 
                #bitin = bin(int(in_column))[2:].zfill(self.outbit)
                check = 1
                for i in range(self.outbit - self.initbit): 
                    if in_column%2 == 1:
                        #out1 = out1 ^ (pow(2,self.outbit - i -1))
                        out1 = out1 ^ check
                        break
                    in_column = in_column >> 1
                    check = check << 1
                return out1
            
    #delete the first row of each block
    def encode_symbol(self, in_symbol):
        while(1):
            out = self.encode_symbol_rough(in_symbol)
            if out%self.partsz != 0:
                return out
    
    def encode_string(self,in_list):  # encode string into a privatized string
        out_list = [self.encode_symbol(self.permute[x]) for x in in_list]
        return out_list
    
    
    def decode_string(self, out_list, iffast = 1, normalization = 0): # get the privatized string and learn the original distribution
        #normalization options: 0: clip and normalize(default)
        #                       1: simplex projection
        #                       else: no nomalization
        
        #iffast: 0 use fast hadamard transform time O(n  + k\log k)
        #        1 no fast hadamard tansform  time O(n + k^2)
        
        l = len(out_list)
        count,edges = np.histogram(out_list,range(self.outsz+1))
        freq = count/float(l)
        
        
        if iffast == 1:
            #parts = self.insz//(self.partsz-1) 
            freq_S = np.zeros(self.outsz)
            freq_block = np.zeros(self.partnum)
            for i in range(0, self.partnum):
                Trans = FWHT_A(self.partsz, freq[i*self.partsz: (i+1)*self.partsz])
                freq_block[i] = Trans[0]
                freq_S[i*(self.partsz-1): (i+1)*(self.partsz-1)] = ( - Trans[1:self.partsz] + Trans[0])/float(2)         
            dist_S = freq_S[0:self.insz]
            
        else:
            freq_block = np.zeros(self.part) # count the number of appearances of each block
            for i in range(0,self.part): 
                #count_block[i] = np.sum(count[i*self.partsz : (i+1)*self.partsz - 1])
                for j in range(0,self.partsz):
                    freq_block[i] = freq_block[i] + freq[i*self.partsz+j]
            #freq_block = np.true_divide(count_block,l) # calculate the frequency of each block
            #dist_block = np.true_divide((2*self.part-1+self.exp)*(freq_block)-2,self.exp-1) # calculate the estimated original prob of each block                    
            for i in range(0, self.insz): 
                pi = int(i)//(self.partsz-1)
                ti = pi*self.partsz + int(i)%(self.partsz-1)+1
                for x in range(pi*self.partsz, (pi+1)*self.partsz): # count the number of appearances of each C_i
                    if self.entry_check(ti,x) == 0:
                        dist_S[i] = dist_S[i] + freq[x]
                        
        lbd = float(self.outsz - self.partnum)/float(self.num_one)
        c1 = lbd-1+self.exp
        
        dist_block = np.true_divide(c1*(freq_block)- 2 + 1/float(self.num_one),self.exp-1) # calculate the estimated original prob of each block
        
        c2 = self.exp - 1
        #dist = [float(2*c1*dist_S[i] - c2*dist_block[i//(self.partsz-1)] - 2)/float(c3) for i in range(0,self.insz) ]
        dist = [float(2*c1*dist_S[i] - c2*dist_block[i//(self.partsz-1)] - 2)/float(c2) for i in range(0,self.insz) ]
        
        if normalization == 0: 
            dist = probability_normalize(dist) #clip and normalize
        if normalization == 1:
            dist = project_probability_simplex(dist) #simplex projection
        
        #reverse the permuation
        dist1 = np.zeros(self.insz)
        for i in range(self.insz):
            dist1[int(self.reverse[i])] = dist[i]
        return dist1
    
    
    def decode_string_old(self, out_list): # get the privatized string and learn the original distribution
        
        l = len(out_list)
        dist_S = np.zeros(self.insz)
        count,edges = np.histogram(out_list,range(self.outsz+1))
        freq = count/float(l)
        
        freq_block = np.zeros(self.part) # count the number of appearances of each block
        for i in range(0,self.part): 
            #count_block[i] = np.sum(count[i*self.partsz : (i+1)*self.partsz - 1])
            for j in range(0,self.partsz):
                freq_block[i] = freq_block[i] + freq[i*self.partsz+j]
        
        
        #freq_block = np.true_divide(count_block,l) # calculate the frequency of each block
        dist_block = np.true_divide((2*self.part-1+self.exp)*(freq_block)-2,self.exp-1) # calculate the estimated original prob of each block
                    
        for i in range(0, self.insz): 
            pi = int(i)//(self.partsz-1)
            ti = pi*self.partsz + int(i)%(self.partsz-1)+1
            for x in range(pi*self.partsz, (pi+1)*self.partsz): # count the number of appearances of each C_i
                if self.entry_check(ti,x) == 0:
                    dist_S[i] = dist_S[i] + freq[x]

        #dist_S = np.zeros(self.insz)
        #dist_S = np.true_divide(num,l) #calculate the frequency of each C_i
        dist_inter = np.true_divide(2*(dist_S*(2*self.part-1+self.exp)-1),self.exp-1) # calculate intermediate prob
        dist = [dist_inter[i] - dist_block[i//(self.partsz-1)] for i in range(0,self.insz)] # calculate the estimated prob for each symbol
        dist = np.maximum(dist,0) #map it to be positive
        norm = np.sum(dist)
        dist = np.true_divide(dist,norm) #ensure the l_1 norm is one
        return dist
    
    #def decode_string_normalize(self, out_list): #normalized outputs using clip and normalize
    #    dist = self.decode_string_permute(out_list)
    #    dist = probability_normalize(dist)
    #    return dist
    
    #def decode_string_project(self, out_list): #projected outputs
    #    dist = self.decode_string_permute(out_list)
    #    dist = project_probability_simplex(dist)
    #    return dist
    
    #def decode_string_permute(self, out_list): # get the privatized string and learn the original distribution
    #    dist1 = self.decode_string_fast(out_list)
    #    dist = np.zeros(self.insz)
    #    for i in range(self.insz):
    #        dist[int(self.reverse[i])] = dist1[i]
    #    return dist


# In[ ]:


#The Hadamard randomized response for all regimes (original version)
class Hadamard_Rand_general_original:
    def __init__(self, absz, pri_para, encode_acc = 0): # absz: alphabet size, pri_para: privacy parameter
        #set encode_acc = 1 to enable fast encoding by intializing hadamard matrix
        
        self.insz = absz #input alphabet size k
        #self.outsz = int(math.pow(2,math.ceil(math.log(absz+1,2)))) #output alphabet size: smallest exponent of 2 which is bigger than k
        #self.outbit = int(math.ceil(math.log(absz+1,2))) #output bit length
        self.pri_para = 1/(1+math.exp(pri_para)) #flipping probability to maintain local privacy
        self.exp = math.exp(pri_para) #privacy parameter
        #self.initbit = int(math.floor(math.log(self.exp,2))) # number of bits indicating the block number 
        self.initbit = int(math.floor(math.log(min(2*absz,self.exp),2))) # number of bits indicating the block number
        self.part = int(math.pow(2,self.initbit)) #number of blocks B
        self.tailbit = int(math.ceil(math.log(float(self.insz)/float(self.part)+1,2))) #number of bits indicating the location within a block
        self.partsz = int(math.pow(2,self.tailbit)) # size of each block b
        self.outbit = self.tailbit + self.initbit #total number of output bits
        self.outsz = int(math.pow(2,self.outbit)) # output alphabet size K
        self.permute, self.reverse = Random_Permutation(absz) #Initialize random permutation
        
        self.ifencode_acc = encode_acc #whether to accelerate encoding process
        if encode_acc == 1:
            self.H = Hadarmard_init(self.partsz) # initialize Hadarmard matrix
        
        
    def entry_check(self,x,y): #check if the reduced hadamard entry is one (return 0 for 1)
        x_bit = bin(x)[2:].zfill(self.outbit)
        y_bit = bin(y)[2:].zfill(self.outbit)
        for i in range(0,self.initbit): # check if they are in the same block, if not, return -1
            if x_bit[i] != y_bit[i]:
                return True
        check = False
        for i in range(self.initbit, self.outbit): #check whether the entry is one within the block
            check = check^(int(x_bit[i]) & int(y_bit[i]))
        return check
                                  
    
    def encode_symbol(self,in_symbol):  # encode a single symbol into a privatized version 
        # we use coupling argument to do this
        part_index = int(in_symbol)//(self.partsz-1)
        part_pos = int(in_symbol)%(self.partsz-1)+1
        in_column = (part_index << self.tailbit) + part_pos #map the input x to the xth column with weight b/2
        #in_column = part_index * self.partsz + part_pos
        out1 = np.random.randint(0,self.outsz) #get a random number out1 in the output alphabet as a potential output
        ra = random.random()
        if ra < (2*self.part)/(2*self.part-1+self.exp): #with certain prob, output the same symbol as from uniform distribution
            return out1
        else:
            out_pos = out1 & (self.partsz - 1)
            #out_pos = out1 % self.partsz
            out1 =  out_pos + (part_index << self.tailbit) # map out1 to the same block as in_column while maintain the location within the block
            #out1 = out_pos + part_index*self.partsz
            if self.ifencode_acc == 1:
                check = 1 - self.H[part_pos][out_pos]
            else:
                check = self.entry_check(in_column,out1)

            if check == 0: #if location (in_column, out1) is one, output out1
                return out1
            else: #else flip bit at the left most location where bit representation of in_column is one 
                #bitin = bin(int(in_column))[2:].zfill(self.outbit)
                check = 1
                for i in range(self.outbit - self.initbit): 
                    if in_column%2 == 1:
                        #out1 = out1 ^ (pow(2,self.outbit - i -1))
                        out1 = out1 ^ check
                        break
                    in_column = in_column >> 1
                    check = check << 1
                return out1
    
    def encode_string(self,in_list): #permute before encoding
        out_list = [self.encode_symbol(self.permute[x]) for x in in_list]
        return out_list        
    
    
    def decode_string(self, out_list,iffast = 1, normalization = 0): # get the privatized string and learn the original distribution
        #normalization options: 0: clip and normalize(default)
        #                       1: simplex projection
        #                       else: no nomalization
        
        #iffast: 0 use fast hadamard transform time O(n  + k\log k)
        #        1 no fast hadamard tansform  time O(n + kB), B is the block size
        
        l = len(out_list)
        count,edges = np.histogram(out_list,range(self.outsz+1))
        freq = count/float(l)
        
        if iffast == 1:
            parts = self.insz//(self.partsz-1) 
            freq_S = np.zeros((parts+1)*self.partsz)
            freq_block = np.zeros((parts+1)*self.partsz)
        
            for i in range(0, parts+1):
                Trans = FWHT_A(self.partsz, freq[i*self.partsz: (i+1)*self.partsz])
                freq_block[i] = Trans[0]
                freq_S[i*(self.partsz-1): (i+1)*(self.partsz-1)] = (Trans[1:self.partsz] + Trans[0])/float(2) 
            dist_S = freq_S[0:self.insz]
        
            dist_block = np.true_divide((2*self.part-1+self.exp)*(freq_block)-2,self.exp-1) # calculate the estimated original prob of each block
        
        else:
            freq_block = np.zeros(self.part) # count the number of appearances of each block
            for i in range(0,self.part): 
                #count_block[i] = np.sum(count[i*self.partsz : (i+1)*self.partsz - 1])
                for j in range(0,self.partsz):
                    freq_block[i] = freq_block[i] + freq[i*self.partsz+j]     
                    
            dist_block = np.true_divide((2*self.part-1+self.exp)*(freq_block)-2,self.exp-1) # calculate the estimated original prob of each block
            for i in range(0, self.insz): 
                pi = int(i)//(self.partsz-1)
                ti = pi*self.partsz + int(i)%(self.partsz-1)+1
                for x in range(pi*self.partsz, (pi+1)*self.partsz): # count the number of appearances of each C_i
                    if self.entry_check(ti,x) == 0:
                        dist_S[i] = dist_S[i] + freq[x]
        
        dist_inter = np.true_divide(2*(dist_S*(2*self.part-1+self.exp)-1),self.exp-1) # calculate intermediate prob
        dist = [dist_inter[i] - dist_block[i//(self.partsz-1)] for i in range(0,self.insz)] # calculate the estimated prob for each symbol
        
        if normalization == 0: 
            dist = probability_normalize(dist) #clip and normalize
        if normalization == 1:
            dist = project_probability_simplex(dist) #simplex projection
        
        #reverse the permuation
        dist1 = np.zeros(self.insz)
        for i in range(self.insz):
            dist1[int(self.reverse[i])] = dist[i]
        return dist1



import math
import xxhash
from pure_ldp.core import FreqOracleServer

# Server-side for local-hashing

# Very loosely based on code by Wang (https://github.com/vvv214/LDP_Protocols/blob/master/olh.py)

class LHServer(FreqOracleServer):
    def __init__(self, epsilon, d, g=2, use_olh=False, index_mapper=None):
        """

        Args:
            epsilon: float - The privacy budget
            d: integer - Size of the data domain
            g: Optional float - The domain [g] = {1,2,...,g} that data is hashed to, 2 by default (binary local hashing)
            use_olh: Optional boolean - if set to true uses Optimised Local Hashing i.e g is set to round(e^epsilon + 1)
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.set_name("LHServer")
        self.g = g
        self.use_olh = use_olh
        self.update_params(epsilon, d, g, index_mapper)

    def update_params(self, epsilon=None, d=None, g=None, index_mapper=None):
        """
        Updates LHServer parameters, will reset any aggregated/estimated data
        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            g: optional - hash domain
            index_mapper: optional - function
        """
        super().update_params(epsilon, d, index_mapper)

        # Update probs and g
        if epsilon is not None:
            if self.use_olh is True:
                self.g = int(round(math.exp(self.epsilon))) + 1

            self.p = math.exp(self.epsilon) / (math.exp(self.epsilon) + self.g - 1)

    def aggregate(self, priv_data):
        """
        Aggregates privatised data from LHClient to be used to calculate frequency estimates.

        Args:
            priv_data: Privatised data of the form returned from UEClient.privatise
            seed: kwarg - The seed of the user's hash function, must be passed as a keyword arg
        """
        seed = priv_data[1]
        priv_data = priv_data[0]

        for i in range(0, self.d):
            if priv_data == (xxhash.xxh32(str(i), seed=seed).intdigest() % self.g):
                self.aggregated_data[i] += 1

        self.n += 1

    def _update_estimates(self):
        a = self.g / (self.p * self.g - 1)
        b = self.n / (self.p * self.g - 1)

        self.estimated_data = a * self.aggregated_data - b
        return self.estimated_data

    def estimate(self, data, suppress_warnings=False):
        """
        Calculates a frequency estimate of the given data item using the aggregated data.

        Args:
            data: data item
            suppress_warnings: Optional boolean - Suppresses warnings about possible inaccurate estimations

        Returns: float - frequency estimate of the data item

        """
        self.check_warnings(suppress_warnings=suppress_warnings)
        index = self.index_mapper(data)
        self.check_and_update_estimates()
        return self.estimated_data[index]



from pure_ldp.core import FreqOracleServer
from pure_ldp.frequency_oracles.hadamard_response.internal import k2k_hadamard
import math


class HadamardResponseServer(FreqOracleServer):
    def __init__(self, epsilon, d, index_mapper=None, normalization=0):
        """

        Args:
            epsilon (float): Privacy Budget
            d (int): Domain size
            index_mapper (Optional function): A function that maps domain elements to {0, ... d-1}
            normalisation (Optional int): 0 (default) - No normalisation
                           1 - Normalisation (+ clip to 0)
                           2 - Projects estimates onto the probability simplex
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.aggregated_data = []
        self.update_params(epsilon, d, index_mapper)
        self.set_name("Hadamard Response")
        self.normalization = normalization

    def get_hash_funcs(self):
        if self.epsilon > 1:
            return self.hr.permute
        else:
            return

    def reset(self):
        """
        Resets aggregated/estimated data to allow for new collection/aggregation
        """
        super().reset()
        self.aggregated_data = [] # For Hadamard response, aggregated data is stored in a list not numpy array

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Updates HR Server parameters, will reset aggregated/estimated data.
        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            index_mapper: optional - function
        """
        super().update_params(epsilon, d, index_mapper)
        if d is not None or epsilon is not None:
            if self.epsilon <= 1:
                self.hr = k2k_hadamard.Hadamard_Rand_high_priv(d, self.epsilon, encode_acc=1) # hadamard_response
            else:
                self.hr = k2k_hadamard.Hadamard_Rand_general_original(d, self.epsilon, encode_acc=1)

    def aggregate(self, data):
        """
        Used to aggregate privatised data to the server
        Args:
            data: privatised data item to aggregate
        """
        self.aggregated_data.append(data)
        self.n += 1

    def _update_estimates(self):
        """
        Used internally to update estimates
        Returns: estimated data

        """
        self.estimated_data = self.hr.decode_string(self.aggregated_data, normalization=self.normalization-1) * self.n # k2khadamard using norm=0 for normalisation, 1 for simplex and anything else for none
        return self.estimated_data

    def estimate(self, data, suppress_warnings=False):
        """
        Estimates frequency of given data item
        Args:
            data: data item to estimate
            suppress_warnings: Optional boolean - If True, estimation warnings will not be displayed

        Returns: frequency estimate

        """
        self.check_warnings(suppress_warnings)
        index = self.index_mapper(data)
        self.check_and_update_estimates()
        return self.estimated_data[index]



import warnings
import numpy as np

class FreqOracleServer:
    def __init__(self, epsilon, d, index_mapper=None):
        """

        Args:
            epsilon: privacy budget
            d: domain size - not all freq oracles need this so can be None
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain

        """
        self.epsilon = epsilon
        self.d = d

        self.aggregated_data = np.zeros(self.d) # Some freq oracle servers keep track of aggregated data to generate estimated_data
        self.estimated_data = [] # Keep track of estimated data for quick access
        self.n = 0 # The number of data items aggregated

        self.name = "FrequencyOracle" # Name of the frequency oracle for warning messages, set using .set_name(name)
        self.last_estimated = 0

        if index_mapper is None:
            self.index_mapper = lambda x: x - 1
        else:
            self.index_mapper = index_mapper

    def set_name(self, name):
        """
        Set's freq servers name
        Args:
            name: string - name of frequency oracle
        """
        self.name = name

    def reset(self):
        """
        This method resets the server's aggregated/estimated data and sets n = 0.
        This should be overridden if other parameters need to be reset.
        """
        self.aggregated_data = np.zeros(self.d)
        self.estimated_data = []
        self.last_estimated = 0
        self.n = 0

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Method to update params of freq oracle server, should be overridden if more options needed.
        This will reset aggregated/estimated data.
        Args:
            epsilon: Optional - privacy budget
            d: Optional - domain size
            index_mapper: Optional - function
        """
        self.epsilon = epsilon if epsilon is not None else self.epsilon # Updating epsilon here will not update any internal probabilities
        # Any class that implements FreqOracleServer, needs to override update_params to update epsilon properly

        self.d = d if d is not None else self.d
        self.index_mapper = index_mapper if index_mapper is not None else self.index_mapper
        self.reset()

    def check_warnings(self, suppress_warnings=False):
        """
        Used during estimation to check warnings
        Args:
            suppress_warnings: Optional boolean - If True suppresses warnings from being output
        """
        if not suppress_warnings:
            if self.n < 10000:
                warnings.warn(self.name + " has only aggregated small amounts of data (n=" + str(self.n) +
                              ") estimations may be highly inaccurate", RuntimeWarning)
            if self.epsilon < 1:
                warnings.warn("High privacy has been detected (epsilon = " + str(self.epsilon) +
                              "), estimations may be highly inaccurate on small datasets", RuntimeWarning)

    def aggregate(self, data):
        """
        The main method for aggregation, should be implemented by a freq oracle server
        Args:
            data: item to estimate frequency of
        """
        raise NotImplementedError("Must implement")

    def aggregate_all(self, data_list):
        """
        Helper method used to aggregate a list of data
        Args:
            data_list: List of private data to aggregate
        """
        for data in data_list:
            self.aggregate(data)

    def check_and_update_estimates(self):
        """
        Used to check if the "cached" estimated data needs re-estimating, this occurs when new data has been aggregated since last
        """
        if self.last_estimated < self.n:  # If new data has been aggregated since the last estimation, then estimate all
            self.last_estimated = self.n
            self._update_estimates()

    def _update_estimates(self):
        """
        Used internally to update estimates, should be implemented
        """
        raise NotImplementedError("Must implement")

    def estimate(self, data, suppress_warnings=False):
        """
        Calculates frequency estimate of given data item, must be implemented
        Args:
            data: data to estimate the frequency warning of
            suppress_warnings: Optional boolean - if true suppresses warnings
        """
        raise NotImplementedError("Must implement")

    def estimate_all(self, data_list, suppress_warnings=False):
        """
        Helper method, given a list of data items, returns a list of their estimated frequencies
        Args:
            data_list: list of data items to estimate
            suppress_warnings: If True, will suppress estimation warnings

        Returns: list of estimates

        """
        self.check_and_update_estimates()
        return [self.estimate(x, suppress_warnings=suppress_warnings) for x in data_list]

    @property
    def get_estimates(self):
        """
        Returns: Estimated data
        """
        return self.estimated_data



import numpy as np

from pure_ldp.core import FreqOracleClient

# Client-side for histogram-encoding

class HEClient(FreqOracleClient):
    def __init__(self, epsilon, d, index_mapper=None):
        """

        Args:
            epsilon: float - the privacy budget
            d: integer - the size of the data domain
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)

    def _perturb(self, oh_vec):
        """
        Used internally to peturb data using Laplacian noise following the histogram encoding technique.

        Args:
            oh_vec: A one-hot vector, where the entry of the user's data item is set to 1 and everything else 0

        Returns: a privatised noisy vector

        """
        noise = np.random.laplace(scale=(2 / self.epsilon), size=self.d)
        noisy_vec = oh_vec + noise
        return noisy_vec

    def privatise(self, data):
        """
        Used to privatise a user's data using histogram encoding (Laplace mechanism)

        Args:
            data: User's data item

        Returns: Privatised data vector

        """
        index = self.index_mapper(data)

        oh_vec = np.zeros(self.d)
        oh_vec[index] = 1

        return self._perturb(oh_vec)



from pure_ldp.core import FreqOracleClient

import numpy as np
from scipy.linalg import hadamard
import math
import random

class CMSClient(FreqOracleClient):
    def __init__(self, epsilon, hash_funcs, m, is_hadamard=False):
        """
        Apple's Count Mean Sketch (CMS) Algorithm
        Args:
            epsilon (float): Privacy Budget Epsilon
            hash_funcs (list of funcs): A list of hash function mapping data to {0...m-1} (can be generated by CMSServer)
            m (int): The length of the hash domain
            is_hadamard (optional bool): If true, uses Hadamard Count Mean Sketch (HCMS)
        """
        super().__init__(epsilon, None)
        self.update_params(hash_funcs, m, epsilon)
        self.is_hadamard = is_hadamard

        if self.is_hadamard:
            self.had = hadamard(self.m)

    def update_params(self, hash_funcs=None, m=None, epsilon=None, index_mapper=None):
        """
        Updates parameters
        Args:
            hash_funcs (optional list): List of k hash functions mapping data to {0...m-1}
            m (optional int): Length of hash domain
            epsilon (optional int): Privacy Budget
            index_mapper (optional function): Index mapper function
        """
        if hash_funcs is not None:
            self.hash_funcs = hash_funcs
            self.k = len(self.hash_funcs)

        self.epsilon = epsilon if epsilon is not None else self.epsilon
        self.m = m if m is not None else self.m

        if epsilon is not None:
            self.prob = 1 / (1 + math.pow(math.e, self.epsilon / 2))

    def _one_hot(self, data):
        """
        Used internally to perturb data
        Args:
            data: arbitrary data

        Returns: perturbed vector v and hash index j

        """
        j = random.randint(0, self.k-1)
        h_j = self.hash_funcs[j]
        v = [0] * self.m if self.is_hadamard else np.full(self.m, -1)
        v[h_j(data)] = 1
        return v, j

    def _cms_perturb(self, data):
        """
        Used internally for peturbing data using the CMS algorithm
        Args:
            data: data to be perturbed

        Returns: peturbed cms data

        """
        v, j = self._one_hot(data)
        v[np.random.rand(*v.shape) < self.prob] *= -1 # "flip" bits with prob
        # return np.multiply(v, b), j # Used to generate a random vector b using np.random.choice but it was 3x slower than the above line
        return v, j

    def _hcms_perturb(self, data):
        """
        Used internally for perturbing data using HCMS
        Args:
            data: data to be perturbed

        Returns: peturbed hcms data

        """
        if not (self.m & (self.m - 1)) == 0:
            raise ValueError("m must be a positive integer, and m must be a power of 2 to use hcms")

        v, j = self._one_hot(data)
        b = random.choices([-1, 1], k=1, weights=[self.prob, 1 - self.prob])
        h_j = self.hash_funcs[j]
        w = self.had[:, h_j(data)]
        l = random.randint(0, self.m-1)
        return b[0] * w[l], j, l  # Return (b*w_l, index j, index l)

    def privatise(self, data):
        """
        Privatises data item using CMS/HCMS

        Args:
            data: item to be privatised

        Returns: Privatised data

        """
        data = str(data)
        if self.is_hadamard:
            return self._hcms_perturb(data)
        else:
            return self._cms_perturb(data)


import math
from scipy.optimize import fminbound
from pure_ldp.core import FreqOracleServer

# Client-side for histogram-encoding

class HEServer(FreqOracleServer):
    def __init__(self, epsilon, d, use_the=False, theta=None, index_mapper=None):
        """

        Args:
            epsilon: float - the privacy budget
            d: integer - the size of the data domain
            use_the: Optional boolean - If set to true uses Thresholding Histogram Encoding (THE)
            theta: Optional - If passed, will override the optimal theta value (not recommended)
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """

        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.set_name("HEServer")

        self.is_the = use_the

        if use_the is True:
            self.theta = theta
            if theta is None:
                self.theta = self.__find_optimal_theta()

            self.p = 1 - 0.5 * (math.pow(math.e, self.epsilon / 2 * (self.theta - 1)))
            self.q = 0.5 * (math.pow(math.e, -1 * self.theta * (self.epsilon / 2)))

    def update_params(self, epsilon=None, d=None, theta=None, index_mapper=None):
        super().update_params(epsilon, d, index_mapper)

        if self.is_the:
            self.theta = theta
            if theta is None:
                self.theta = self.__find_optimal_theta()

            self.p = 1 - 0.5 * (math.pow(math.e, self.epsilon / 2 * (self.theta - 1)))
            self.q = 0.5 * (math.pow(math.e, -1 * self.theta * (self.epsilon / 2)))

    def __find_optimal_theta(self):
        """
        Used internally to calculate the optimal value of theta if using Threshold Histogram Encoding (THE).

        Find the optimal value of theta to use for thresholding, by minimising
        the variance equation for the fixed value of epsilon.

        Returns: float to 4dp - optimal theta parameter

        """
        # Minimise variance for our fixed epsilon to find the optimal theta value for thresholding
        def var(x):
            num = 2 * math.exp(self.epsilon * x / 2) - 1
            denom = math.pow(1 + math.exp((self.epsilon * (x - 1/2))) - 2 * math.exp((self.epsilon * x / 2)), 2)
            return num/denom

        return round(float(fminbound(var, 0.5, 1)), 4)

    def aggregate(self, priv_data):
        """
        Aggregates HE privatised data, to allow us to calculate estimates.

        Args:
            priv_data: Data privatised via he_client.privatise
        """
        # Threshold rounding
        if self.is_the:
            for index, item in enumerate(priv_data):
                if item <= self.theta:
                    priv_data[index] = 0
                else:
                    priv_data[index] = 1

        self.aggregated_data += priv_data
        self.n += 1

    def _update_estimates(self):
        self.estimated_data = self.aggregated_data
        if self.is_the:
            self.estimated_data = (self.aggregated_data - self.n * self.q) / (self.p - self.q)
        return self.estimated_data

    def estimate(self, data, suppress_warnings=False):
        """
        Calculates a frequency estimate of the given data item

        Args:
            data: data item
            suppress_warnings: Optional boolean - Supresses warnings about possible inaccurate estimations

        Returns: float - frequency estimate

        """
        self.check_warnings(suppress_warnings=suppress_warnings)
        index = self.index_mapper(data)
        self.check_and_update_estimates()
        return self.estimated_data[index]



from pure_ldp.frequency_oracles.local_hashing import LHServer
from pure_ldp.core import FreqOracleServer

import math
import itertools
import copy
import numpy as np

from bitstring import BitArray
from collections import Counter


class PEMServer:
    def __init__(self, epsilon, domain_size, start_length, segment_length, FOServer=None):
        """

        Args:
            epsilon: float privacy budget
            domain_size: max string length
            start_length: starting size of the fragment
            segment_length: length to increase fragment by on each round
            FOServer: instance of FreqOracleServer to aggregate and estimate the heavy hitters
        """
        self.epsilon = epsilon
        self.domain_size = domain_size
        self.segment_length = segment_length
        self.start_length = start_length

        self.g = math.ceil((self.domain_size - self.start_length) / self.segment_length)
        self.oracles = []
        self.n = 0

        if isinstance(FOServer, FreqOracleServer):
            for i in range(0, self.g):
                oracle = copy.deepcopy(FOServer)
                d = 2 ** (self.start_length + (i + 1) * self.segment_length)
                oracle.update_params(d=d,
                                     index_mapper=lambda x: x) # Some oracles need a domain size
                oracle.reset()
                self.oracles.append(oracle)
        else:
            for i in range(0, self.g):
                self.oracles.append(
                    LHServer(self.epsilon, 2 ** (self.start_length + (i + 1) * self.segment_length), use_olh=True, index_mapper= lambda x:x))

    def aggregate(self, pem_data):
        """

        Args:
            privatised_fragment: a privatised bit string from PEMClient
            group: the group number
        """
        privatised_fragment, group = pem_data
        self.oracles[group].aggregate(privatised_fragment)
        self.n += 1

    def _estimate_top_k(self, oracle, candidates, k):
        """

        Args:
            oracle: frequncy oracle (FreqOracleServer instance)
            candidates: a list of candidate strings
            k: int - used to find the top k most frequent strings

        Returns:

        """
        # TODO: Faster/nicer way to do this?
        top_k, _ = zip(*Counter(dict(zip(candidates, oracle.estimate_all(candidates, suppress_warnings=True)))).most_common(k))
        return top_k

    def find_top_k(self, k):
        """

        Args:
            k: int - used to find the top k most frequent strings

        Returns: list of top-k frequent candidates, and a list of there estimated frequencies

        """
        fragment_size = self.start_length + (0 + 1) * self.segment_length
        candidates = range(0, 2 ** fragment_size)
        top_k = self._estimate_top_k(self.oracles[0], candidates, k)

        freq_candidates = list(map(lambda x: BitArray(uint=x, length=fragment_size).bin, top_k))

        for i in range(1, self.g):
            fragment_size = self.start_length + (i + 1) * self.segment_length

            frags = [''.join(comb) for comb in itertools.product(["0", "1"], repeat=self.segment_length)]

            candidates = []
            for frag in frags:
                candidates.extend([BitArray(bin=bs + frag).uint for bs in freq_candidates])

            top_k = self._estimate_top_k(self.oracles[i], candidates, k)

            freq_candidates = list(map(lambda x: BitArray(uint=x, length=fragment_size).bin, top_k))

        freqs = self.g * np.array(self.oracles[self.g-1].estimate_all([BitArray(bin=x).uint for x in freq_candidates], suppress_warnings=True))
        return freq_candidates, freqs



from pure_ldp.heavy_hitters import *
import copy
import inspect


# Used to create a list of possible frequency oracles in the pure-LDP library

client_class_list = []
server_class_list = []
globs = list(globals().keys()).copy()  # Create copy, since globals updates too much to iterate through

for key in globs:
    if "Client" in key:
        client_class_list.append(key.replace("Client", ""))
    if "Server" in key:
        server_class_list.append(key.replace("Server", ""))

class_list = {"Client": client_class_list, "Server": server_class_list}


def _create_hh_instance(obj_type, name, obj_params):
    """
    Used internally to create a heavy hitter instance (client/server)

    Args:
        obj_type: Either "Client" or "Server"
        name: Name of heavy hitter algorithm
        obj_params: Parameters for the client/server object

    Returns: name + obj_type heavy hitter object

    """
    fo_list = class_list[obj_type]

    split = name.split("_")  # Get prefix of client name i.e if passed "local_hashing" get "LH" as prefix

    if len(split) > 1:
        name = ""
        for word in split:
            name += word[0]

    name = name.upper()
    if name == "HR": name = "HADAMARDRESPONSE"

    upper_hh_list = list(map(lambda x: x.upper(), fo_list))

    if name not in upper_hh_list:
        raise ValueError("Heavy Hitter", obj_type, "must be one of:", fo_list,
                         "\n NOTE: Values are case insensitive")

    hh_name = client_class_list[upper_hh_list.index(name)] + obj_type

    constructor = globals().get(hh_name)
    expected_params = list(inspect.signature(constructor).parameters)

    params = dict(
        (key.split("=")[0], obj_params[key.split("=")[0]]) for key in expected_params if key in obj_params.keys())

    return constructor(**params)


def create_hh_client_instance(name, client_params):
    """
    Returns a heavy hitter client given the name and parameters

    Args:
        name: Name of heavy hitter algorithm (i.e PEM)
        client_params: The parameters for the client

    Returns: nameClient heavy hitter object

    """
    return _create_hh_instance("Client", name, client_params)


def create_hh_server_instance(name, server_params):
    """
    Returns a heavy hitter server given the name and parameters

    Args:
        name: Name of heavy hitter algorithm (i.e PEM)
        server_params: The parameters for the server

    Returns: nameServer heavy hitter object

    """
    return _create_hh_instance("Server", name, server_params)



import numpy as np
import math

from pure_ldp.core import FreqOracleClient

# Client-side for unary-encoding
    # By default parameters are set for Symmetric Unary Encoding (SUE)
    # If is_oue=True is passed to the constructor then it uses Optimised Unary Encoding (OUE)

class UEClient(FreqOracleClient):
    def __init__(self, epsilon, d, use_oue=False, index_mapper=None):
        """

        Args:
            epsilon: float - privacy budget
            d: integer - the size of the data domain
            use_oue: Optional boolean - if True, will use Optimised Unary Encoding (OUE)
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.use_oue = use_oue
        self.update_params(epsilon, d, index_mapper)

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Used to update the client UE parameters.
        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            index_mapper:  optional - function
        """
        super().update_params(epsilon, d, index_mapper)

        if epsilon is not None: # If epsilon changes, update probs
            const = math.pow(math.e, self.epsilon/2)
            self.p = const / (const + 1)
            self.q = 1-self.p

            if self.use_oue is True:
                self.p = 0.5
                self.q = 1/(math.pow(math.e, self.epsilon) + 1)

    def _perturb(self, index):
        """
        Used internally to peturb data using unary encoding

        Args:
            index: the index corresponding to the data item

        Returns: privatised data vector

        """
        oh_vec = np.random.choice([1, 0], size=self.d, p=[self.q, 1-self.q])  # If entry is 0, flip with prob q
        oh_vec[index] = np.random.choice([1, 0], p=[self.p, 1-self.p]) # If entry is 1, keep as 1 with prob p
        return oh_vec

    def privatise(self, data):
        """
        Privatises a user's data item using unary encoding.

        Args:
            data: data item

        Returns: privatised data vector

        """
        index = self.index_mapper(data)
        return self._perturb(index)




class FreqOracleClient:
    def __init__(self, epsilon, d, index_mapper=None):
        """

        Args:
            epsilon (float): Privacy budget
            d (int): domain size - not all freq oracles need this, so can be None
            index_mapper (func): Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        self.epsilon = epsilon
        self.d = d

        if index_mapper is None:
            self.index_mapper = lambda x: x - 1
        else:
            self.index_mapper = index_mapper

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Method to update params of freq oracle client, should be overridden if more options needed.
        Args:
            epsilon (optional float): Privacy budget
            d (optional int): Domain size
            index_mapper (optional func): Index map function
        """
        self.epsilon = epsilon if epsilon is not None else self.epsilon
        self.d = d if d is not None else self.d
        self.index_mapper = index_mapper if index_mapper is not None else self.index_mapper

    def _perturb(self, data):
        """
        Used internally to perturb raw data, must be implemented by a FreqOracle
        Args:
            data: user's data item
        """
        raise NotImplementedError("Must implement")

    def privatise(self, data):
        """
        Public facing method to privatise user's data
        Args:
            data: user's data item
        """
        raise NotImplementedError("Must implement")


import xxhash
import hashlib
from bitarray import bitarray

# Base classes for frequency oracles
from ._freq_oracle_client import FreqOracleClient
from ._freq_oracle_server import FreqOracleServer

# Helper functions for generating hash funcs

def generate_hash_funcs(k, m):
    """
    Generates k hash functions that map data to the range {0, 1,..., m-1}
    Args:
        k: The number of hash functions
        m: The domain {0,1,...,m-1} that hash func maps too
    Returns: List of k hash functions
    """
    hash_funcs = []
    for i in range(0, k):
        hash_funcs.append(generate_hash(m, i))
    return hash_funcs


def generate_256_hash():
    """

    Returns: A hash function that maps data to {0,1,... 255}

    """
    return lambda data: xxhash.xxh64(data, seed=10).intdigest() % 256


def generate_hash(m, seed):
    """
    Generate a single hash function that maps data to {0, ... ,m-1}
    Args:
        m: int domain to map too
        seed: int the seed for the hash function

    Returns: Hash function

    """
    return lambda data: xxhash.xxh64(str(data), seed=seed).intdigest() % m


def get_sha256_hash_arr(hashId, dataString):
    """
    Used in priv_count_sketch freq oracle for hashing
    Args:
        hashId: seed of the hash
        dataString: data string to hash

    Returns: hashed data as a bitarray

    """
    message = hashlib.sha256()

    message.update((str(hashId) + dataString).encode("utf8"))

    message_in_bytes = message.digest()

    message_in_bit_array = bitarray(endian='little')
    message_in_bit_array.frombytes(message_in_bytes)

    return message_in_bit_array




import numpy as np
import math
#fast Hadamard Transform
def FWHT_A(k, dist):
    if k == 1:
        return dist
    dist1 = dist[0 : k//2]
    dist2 = dist[k//2 : k]
    trans1 = FWHT_A(k//2, dist1)
    trans2 = FWHT_A(k//2, dist2)
    trans = np.concatenate((trans1+ trans2, trans1 - trans2))
    return trans

#simplex projection
def project_probability_simplex(p_estimate):
    
    k = len(p_estimate)  # Infer the size of the alphabet.
    p_estimate_sorted = np.sort(p_estimate)
    p_estimate_sorted[:] = p_estimate_sorted[::-1]
    p_sorted_cumsum = np.cumsum(p_estimate_sorted)
    i = 1
    while i < k:
        if p_estimate_sorted[i] + (1.0 / (i + 1)) * (1 - p_sorted_cumsum[i]) < 0:
            break
        i += 1
    lmd = (1.0 / i) * (1 - p_sorted_cumsum[i - 1])
    return np.maximum(p_estimate + lmd, 0)

#clip and normalize
def probability_normalize(dist):
    dist = np.maximum(dist,0) #map it to be positive
    norm = np.sum(dist)
    dist = np.true_divide(dist,norm) #ensure the l_1 norm is one
    return dist

#generate a random permutation matrix
def Random_Permutation(k):
    permute = np.random.permutation(k)
    reverse = np.zeros(k)
    for i in range(k):
        reverse[int(permute[i])] = i
    return permute,reverse

#pre-calculate Hadamard Matrix
def Hadarmard_init(k):
    H = [None] * k
    for row in range(k):
        H[row] = [True] * k
        
# Initialize Hadamard matrix of order n.
    i1 = 1
    while i1 < k:
        for i2 in range(i1):
            for i3 in range(i1):
                H[i2+i1][i3]    = H[i2][i3]
                H[i2][i3+i1]    = H[i2][i3]
                H[i2+i1][i3+i1] = not H[i2][i3]
        i1 += i1
    return H


# In[ ]:


#functions to generate distributions
def generate_geometric_distribution(k,lbd):
    elements = range(0,k)
    prob = [(1-lbd)*math.pow(lbd,x)/(1-math.pow(lbd,k)) for x in elements] # geometric dist
    return prob

def generate_uniform_distribution(k):
    raw_distribution = [1] * k
    sum_raw = sum(raw_distribution)
    prob = [float(y)/float(sum_raw) for y in raw_distribution]
    return prob

def generate_two_steps_distribution(k):
    raw_distribution = [1] * int(k/2) + [3] * int(k/2)
    sum_raw = sum(raw_distribution)
    prob = [float(y)/float(sum_raw) for y in raw_distribution]
    return prob

def generate_Zipf_distribution(k,lbd):
    raw_distribution = [1/(float(i)**(lbd)) for i in range(1,k+1)]
    sum_raw = sum(raw_distribution)
    prob = [float(y)/float(sum_raw) for y in raw_distribution]
    return prob

def generate_Dirichlet_distribution(k,lbd):  
    raw_distribution = [0] * k
    for i in range(0,k):
        raw_distribution[i] = np.random.gamma(1,1)
    sum_raw = sum(raw_distribution)
    prob = [float(y)/float(sum_raw) for y in raw_distribution]
    return prob




from setuptools import setup, find_packages
import shutil
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

if os.path.exists("./build"):
    shutil.rmtree("./build")
if os.path.exists("./dist"):
    shutil.rmtree("./dist")
if os.path.exists("./pure_ldp.egg-info"):
    shutil.rmtree("./pure_ldp.egg-info")

setup(
    name='pure-ldp',
    version='1.0.6',
    packages=find_packages(exclude=['*development*', "*apple_sf*", "*treehistogram*", "*rappor*",
                                    "*priv_count_sketch*", "*hashtogram*", "*explicit_hist*", ]),
    install_requires=["xxhash", "numpy", "scipy", "bitstring", "bitarray"],
    url='https://github.com/Samuel-Maddock/pure-LDP',
    license='MIT',
    author='Samuel Maddock',
    author_email='samuel-maddock@hotmail.com',
    description='Simple pure LDP frequency oracle implementations',
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from pure_ldp.core import FreqOracleServer
import math
import numpy as np
from scipy.linalg import hadamard

from pure_ldp.core import generate_hash_funcs

class CMSServer(FreqOracleServer):
    def __init__(self, epsilon, k, m, is_hadamard=False, index_mapper=None):
        """
        Server frequency oracle for Apple's Count Mean Sketch (CMS)

        Args:
            epsilon (float): Privacy Budget
            k (int): Number of hash functions
            m (int): Size of the hash domain
            is_hadamard (optional bool): If True, uses Hadamard Count Mean Sketch (HCMS)
            index_mapper (optional func): Index map function
        """
        super().__init__(epsilon, None, index_mapper)
        self.update_params(k,m, epsilon, d=None, index_mapper=None)
        self.hash_funcs = generate_hash_funcs(k,m)
        self.sketch_matrix = np.zeros((self.k, self.m))
        self.transformed_matrix = np.zeros((self.k, self.m))
        self.is_hadamard = is_hadamard

        self.last_estimated = self.n
        self.ones = np.ones(self.m)

        if self.is_hadamard:
            self.had = hadamard(self.m)

    def update_params(self, k=None, m=None, epsilon=None, d=None, index_mapper=None):
        """
        Updated internal parameters
        Args:
            k (optional int): Number of hash functions
            m (optional int): Size of hash domain
            epsilon (optional float): Privacy Budget
            d (optional int): Size of domain
            index_mapper (optional func): Index map function
        """
        self.k = k if k is not None else self.k
        self.m = m if m is not None else self.m
        self.hash_funcs = generate_hash_funcs(self.k,self.m)
        super().update_params(epsilon=epsilon, d=d, index_mapper=d) # This also calls reset() to reset sketch size
        if epsilon is not None:
            self.c = (math.pow(math.e, epsilon / 2) + 1) / (math.pow(math.e, epsilon / 2) - 1)

    def _add_to_cms_sketch(self, data):
        """
        Given privatised data, adds it to the sketch matrics (CMS algorithm)
        Args:
            data: privatised data by CMS
        """
        item, hash_index = data
        self.sketch_matrix[hash_index] = self.sketch_matrix[hash_index] + self.k * ((self.c / 2) * item + 0.5 * self.ones)

    def _add_to_hcms_sketch(self, data):
        """
        Given privatised data, adds it to the sketch matrix (HCMS algorithm)

        Args:
            data: privatisd data by HCMS
        """
        bit_value, j, l = data
        self.sketch_matrix[j][l] = self.sketch_matrix[j][l] + self.k * self.c * bit_value

    def _transform_sketch_matrix(self):
        """
        Transforms the sketch matrix using inverse hadamard (HCMS)
        Returns: Transformed sketch matrix

        """
        return np.matmul(self.sketch_matrix, np.transpose(self.had))

    def _update_estimates(self):
        """
        If using HCMS, transforms the sketch matrix using inverse hadamard
        """
        if self.is_hadamard:
            self.last_estimated = self.n # TODO: Is this needed?
            self.transformed_matrix = self._transform_sketch_matrix()

    def get_hash_funcs(self):
        """
        Returns hash functions for CMSClient

        Returns: list of k hash_funcs

        """
        return self.hash_funcs

    def reset(self):
        """
        Resets sketch matrix (i.e resets all aggregated data)
        """
        super().reset()
        self.sketch_matrix = np.zeros((self.k, self.m))
        self.transformed_matrix = np.zeros((self.k, self.m))

    def aggregate(self, data):
        """
        Aggregated privatised data
        Args:
            data: Data privatised by CMS/HCMS
        """
        if self.is_hadamard:
            self._add_to_hcms_sketch(data)
        else:
            self._add_to_cms_sketch(data)
        self.n += 1

    def estimate(self, data, suppress_warnings=False):
        """
        Estimates the frequency of the data item

        Args:
            data: item to be estimated
            suppress_warnings (optional bool): If True, will suppress estimation warnings

        Returns: Frequency Estimate

        """
        self.check_warnings(suppress_warnings)
        self.check_and_update_estimates()

        # If it's hadamard we need to transform the sketch matrix
            # To prevent this being performance intensive, we only transform if new data has been aggregated since it was last transformed

        sketch = self.sketch_matrix if not self.is_hadamard else self.transformed_matrix

        data = str(data)
        k, m = sketch.shape
        freq_sum = 0
        for i in range(0, k):
            freq_sum += sketch[i][self.hash_funcs[i](data)]

        return (m / (m - 1)) * ((1 / k) * freq_sum - (self.n / m))




from pure_ldp.frequency_oracles.local_hashing.lh_client import LHClient
from pure_ldp.core import FreqOracleClient

import math
import random

from bitstring import BitArray

class PEMClient:
    def __init__(self, epsilon, domain_size, start_length, segment_length, FOClient=None):
        """

        Args:
            epsilon: privacy budget
            domain_size: max size of the strings to find
            start_length: The starting fragment length
            segment_length: the length to increase the fragment by on each iteration
            FOClient: a FreqOracleClient instance, used to privatise the data
        """
        self.epsilon = epsilon
        self.domain_size = domain_size
        self.segment_length = segment_length
        self.start_length = start_length
        self.g = math.ceil((self.domain_size-self.start_length)/self.segment_length)

        self.index_mapper = lambda x:x
        if isinstance(FOClient, FreqOracleClient):
            self.client = FOClient
        else:
            self.client = LHClient(self.epsilon, d=None, use_olh=True)
        self.client.update_params(index_mapper=self.index_mapper)

    def privatise(self, bit_string):
        """
        This method is used to privatise a bit string using PEM
        Args:
            bit_string: The bit string to be privatised

        Returns: Privatised bit string and the group number

        """
        group = random.randint(0,self.g-1)

        d = 2 ** (self.start_length + (group + 1) * self.segment_length)
        self.client.update_params(d=d)

        fragment_size = self.start_length + (group+1)*self.segment_length
        fragment = bit_string[0:min(fragment_size, len(bit_string))]
        num = BitArray(bin=fragment).uint
        return self.client.privatise(num), group



from pure_ldp.frequency_oracles import *
import copy
import inspect


# Used to create a list of possible frequency oracles in the pure-LDP library

client_class_list = []
server_class_list = []
globs = list(globals().keys()).copy()  # Create copy, since globals updates too much to iterate through

for key in globs:
    if "Client" in key:
        client_class_list.append(key.replace("Client", ""))
    if "Server" in key:
        server_class_list.append(key.replace("Server", ""))

class_list = {"Client": client_class_list, "Server": server_class_list}


def _create_pure_fo_instance(obj_type, name, obj_params):
    """
    Used internally to create instances of various Client/Server frequency oracles

    Args:
        obj_type (str): Either "Client" or "Server"
        name: Name of the frequency oracle to create
        obj_params: Parameters for the frequency oracle object

    Returns: Instance of name + obj_type frequency oracle

    """
    fo_list = class_list[obj_type]

    split = name.split("_")  # Get prefix of client name i.e if passed "local_hashing" get "LH" as prefix

    if len(split) > 1:
        name = ""
        for word in split:
            name += word[0]

    name = name.upper()
    if name == "HR": name = "HADAMARDRESPONSE"

    upper_fo_list = list(map(lambda x: x.upper(), fo_list))

    if name not in upper_fo_list:
        raise ValueError("Frequency oracle must be one of:", fo_list,
                         "\n NOTE: Values are case insensitive")

    fo_name = client_class_list[upper_fo_list.index(name)] + obj_type

    constructor = globals().get(fo_name)
    expected_params = list(inspect.signature(constructor).parameters)

    params = dict(
        (key.split("=")[0], obj_params[key.split("=")[0]]) for key in expected_params if key in obj_params.keys())

    return constructor(**params)


def create_fo_client_instance(name, client_params):
    """
    Given a name of a frequency oracle creates a client instance of it

    Args:
        name: Name of frequency oracle (i.e LH, HE)
        client_params: The parameters for the client frequency oracle object

    Returns: A frequency oracle instance of nameClient

    """
    return _create_pure_fo_instance("Client", name, client_params)


def create_fo_server_instance(name, server_params):
    """
    Given a name of a frequency oracle creates a server instance of it

    Args:
        name: Name of frequency oracle (i.e LH, HE)
        server_params: The parameters for the server frequency oracle

    Returns: A frequency oracle instance of nameServer

    """
    return _create_pure_fo_instance("Server", name, server_params)



import numpy as np
import math
import xxhash
from sys import maxsize
import random
from pure_ldp.core import FreqOracleClient

# Client-side for local-hashing

# Very loosely based on code by Wang (https://github.com/vvv214/LDP_Protocols/blob/master/olh.py)

class LHClient(FreqOracleClient):
    def __init__(self, epsilon, d, g=2, use_olh=False, index_mapper=None):
        """

        Args:
            epsilon: float - The privacy budget
            g: Optional integer - The domain [g] = {1,2,...,g} that data is hashed to, 2 by default (binary local hashing)
            use_olh: Optional boolean - if set to true uses Optimised Local Hashing (OLH) i.e g is set to round(e^epsilon + 1)
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.use_olh = use_olh
        self.g =g
        self.update_params(epsilon, d, g, index_mapper)

    def update_params(self, epsilon=None, d=None, g=None, index_mapper=None):
        """

        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            g: optional - hash domain
            index_mapper: optional - function
        """
        super().update_params(epsilon, d, index_mapper) # Updates core params

        # Updates g and probs
        self.g = g if g is not None else self.g
        if self.use_olh is True:
            self.g = int(round(math.exp(self.epsilon))) + 1

        if self.epsilon is not None or self.g is not None:
            self.p = math.exp(self.epsilon) / (math.exp(self.epsilon) + self.g - 1)
            self.q = 1.0 / (math.exp(self.epsilon) + self.g - 1)

    def _perturb(self, data, seed):
        """
        Used internally to perturb data using local hashing.

        Will hash the user's data item and then peturb it with probabilities that
        satisfy epsilon local differential privacy. Local hashing is explained
        in more detail here: https://www.usenix.org/system/files/conference/usenixsecurity17/sec17-wang-tianhao.pdf

        Args:
            data: User's data to be privatised
            seed: The seed for the user's hash function

        Returns: peturbed data

        """
        index = self.index_mapper(data)

        # Taken directly from Wang (https://github.com/vvv214/LDP_Protocols/blob/master/olh.py#L55-L65)
        x = (xxhash.xxh32(str(index), seed=seed).intdigest() % self.g)
        y = x

        p_sample = np.random.random_sample()
        # the following two are equivalent
        # if p_sample > p:
        #     while not y == x:
        #         y = np.random.randint(0, g)
        if p_sample > self.p - self.q:
            # perturb
            y = np.random.randint(0, self.g)

        return y

    def privatise(self, data):
        """
        Privatises a user's data using local hashing.

        Args:
            data: The data to be privatised

        Returns:
            privatised data: a single integer
        """
        seed = random.randint(0,maxsize) # This is sys.maxsize
        return self._perturb(data, seed), seed



import math

from pure_ldp.core import FreqOracleServer


class UEServer(FreqOracleServer):
    def __init__(self, epsilon, d, use_oue=False, index_mapper=None):
        """
        Args:
            epsilon: float - the privacy budget
            d: integer - the size of the data domain
            use_oue: Optional boolean - If True, will use Optimised Unary Encoding (OUE)
            index_mapper: Optional function - maps data items to indexes in the range {0, 1, ..., d-1} where d is the size of the data domain
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.set_name("UEServer")
        self.use_oue = use_oue
        self.update_params(epsilon, d, index_mapper)

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Updates UE server parameters. This will reset any aggregated/estimated data
        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            index_mapper: optional - index_mapper
        """
        super().update_params(epsilon, d, index_mapper)
        if epsilon is not None:
            const = math.pow(math.e, self.epsilon / 2)
            self.p = const / (const + 1)
            self.q = 1 - self.p

            if self.use_oue is True:
                self.p = 0.5
                self.q = 1 / (math.pow(math.e, self.epsilon) + 1)

    def aggregate(self, priv_data):
        """
        Used to aggregate privatised data by ue_client.privatise

        Args:
            priv_data: privatised data from ue_client.privatise
        """
        self.aggregated_data += priv_data
        self.n += 1

    def _update_estimates(self):
        self.estimated_data = (self.aggregated_data - self.n * self.q) / (self.p - self.q)
        return self.estimated_data

    def estimate(self, data, suppress_warnings=False):
        """
        Calculates a frequency estimate of the given data item

        Args:
            data: data item
            suppress_warnings: Optional boolean - Supresses warnings about possible inaccurate estimations

        Returns: float - frequency estimate

        """
        self.check_warnings(suppress_warnings=suppress_warnings)
        index = self.index_mapper(data)
        self.check_and_update_estimates()
        return self.estimated_data[index]



from pure_ldp.core import FreqOracleClient
from pure_ldp.frequency_oracles.hadamard_response.internal import k2k_hadamard
import math


class HadamardResponseClient(FreqOracleClient):
    def __init__(self, epsilon, d, hash_funcs, index_mapper=None):
        """

        Args:
            epsilon: privacy budget
            d: domain size
            index_mapper: function
        """
        super().__init__(epsilon, d, index_mapper=index_mapper)
        self.update_params(epsilon,d,index_mapper)

        if self.epsilon > 1:
            self.hr.permute = hash_funcs

    def update_params(self, epsilon=None, d=None, index_mapper=None):
        """
        Updates HR client-side params
        Args:
            epsilon: optional - privacy budget
            d: optional - domain size
            index_mapper: optional - function
        """
        super().update_params(epsilon, d, index_mapper)

        if d is not None or epsilon is not None:
            if epsilon <= 1:
                self.hr = k2k_hadamard.Hadamard_Rand_high_priv(d, self.epsilon, encode_acc=1) # hadamard_response
            else:
                self.hr = k2k_hadamard.Hadamard_Rand_general_original(d, self.epsilon, encode_acc=1) # hadamard_response

    def _perturb(self, data):
        """
        Used internally to perturb data
        Args:
            data: item to perturb

        Returns: perturbed data

        """
        if self.epsilon <= 1:
            return self.hr.encode_symbol(data)
        else:
            return self.hr.encode_symbol(self.hr.permute[data]) # General privacy HR needs to permute data

    def privatise(self, data):
        """
        Privatises given data item using the hadamard response technique
        Args:
            data:

        Returns: privatised data

        """
        index = self.index_mapper(data)
        return self._perturb(index)
